import 'package:cloud_firestore/cloud_firestore.dart';

abstract class IdItemInterface {
  String id;
}

abstract class ReadOnlyDatabaseItem<T> extends IdItemInterface {
  Timestamp lastModified;
  String getTypeName();
}

abstract class UpdateableDatabaseItem<T> extends ReadOnlyDatabaseItem<T> {
  T copy();
  Map<String, dynamic> toMap();

  bool equals(UpdateableDatabaseItem<T> item) {
    if (item == null) return false;

    var map = toMap();
    var itemMap = item.toMap();

    if (map.length != itemMap.length) return false;

    for (var key in map.keys) {
      if (!itemMap.containsKey(key)) return false;
      if (key == "last_modified") continue;
      if ((map[key]?.toString() ?? "")
              .compareTo(itemMap[key]?.toString() ?? "") !=
          0) return false;
    }

    return true;
  }
}

class ModelList<T> {
  final DocumentSnapshot snapshot;
  final String fieldName;
  final T Function(DocumentSnapshot snapshot) objectFromSnapshot;

  List<DocumentReference> references;

  ModelList(this.snapshot, this.fieldName, this.objectFromSnapshot);

  List<DocumentReference> getReferences() {
    return references = snapshot.data()[fieldName] == null
        ? <DocumentReference>[]
        : List.from(snapshot.data()[fieldName]);
  }

  List<Future<T>> getObjectList() {
    var list = <Future<T>>[];
    for (var ref in (references ?? getReferences())) {
      list.add(ref.get()?.then(objectFromSnapshot));
    }
    return list;
  }
}
