import 'dart:collection';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class ItemSectionWidget extends StatefulWidget {
  final Widget title;
  final List<Widget> children;

  final IndexedWidgetBuilder itemBuilder;
  final int itemCount;

  final ItemSectionExtendedController controller;

  const ItemSectionWidget({
    Key key,
    this.title,
    this.children,
    this.controller,
  })  : itemBuilder = null,
        itemCount = null,
        super(key: key);

  const ItemSectionWidget.builder({
    Key key,
    this.title,
    @required this.itemBuilder,
    @required this.itemCount,
    this.controller,
  })  : children = null,
        super(key: key);

  @override
  _ItemSectionWidgetState createState() => _ItemSectionWidgetState();
}

class _ItemSectionWidgetState extends State<ItemSectionWidget> {
  bool isExtended;
  bool extendable;

  @override
  void initState() {
    super.initState();
    isExtended = widget.controller?.extended ?? true;
    extendable = widget.controller != null;
  }

  @override
  Widget build(BuildContext context) {
    var columnChildren = <Widget>[];
    if (widget.children != null) {
      var filteredChildren = widget.children
          .where((item) =>
              item != null && !(item is Container && item.child == null))
          .toList();

      filteredChildren.asMap().forEach((index, item) {
        if (item is IListItem)
          columnChildren.add(item.copyWithPosition(
              isFirst: index == 0,
              isLast: index == filteredChildren.length - 1));
        else
          columnChildren.add(item);
      });
    }

    return Column(
      mainAxisSize: MainAxisSize.min,
      crossAxisAlignment: CrossAxisAlignment.start,
      children: <Widget>[
        widget.title != null
            ? Padding(
                padding: const EdgeInsets.symmetric(horizontal: 20),
                child: Material(
                  color: CleanAndroidDesignDefaultColors.background,
                  borderRadius: BorderRadius.circular(6),
                  child: InkWell(
                    onTap: widget.controller == null
                        ? null
                        : () => setState(() {
                              isExtended = !isExtended;
                              widget.controller
                                  ?.setExtendedAndUpdate(isExtended);
                            }),
                    borderRadius: BorderRadius.circular(6),
                    child: Row(
                      mainAxisAlignment: MainAxisAlignment.spaceBetween,
                      children: [
                        widget.title != null
                            ? Padding(
                                padding: const EdgeInsets.symmetric(
                                    horizontal: 5, vertical: 5),
                                child: DefaultTextStyle(
                                  child: widget.title,
                                  style: TextStyles.textStyleSize(14)
                                      .copyWith(fontWeight: FontWeight.w600),
                                ),
                              )
                            : Container(),
                        extendable
                            ? Padding(
                                padding:
                                    const EdgeInsets.symmetric(horizontal: 5),
                                child: Icon(
                                    isExtended
                                        ? Icons.keyboard_arrow_up
                                        : Icons.keyboard_arrow_down,
                                    color: Colors.grey[600]),
                              )
                            : Container(),
                      ],
                    ),
                  ),
                ),
              )
            : Container(),
        isExtended
            ? Container(
                decoration: BoxDecoration(
                    borderRadius: BorderRadius.circular(30),
                    border: Border.all(
                        color:
                            CleanAndroidDesignDefaultColors.defaultColorLight,
                        width: 1),
                    color: Colors.white),
                child: widget.itemBuilder == null
                    ? Column(
                        children: columnChildren,
                      )
                    : ListView.builder(
                        shrinkWrap: true,
                        physics: ScrollPhysics(),
                        itemCount: widget.itemCount,
                        itemBuilder: (context, index) {
                          var widgetChild = widget.itemBuilder(context, index);
                          if (widgetChild is IListItem)
                            return widgetChild.copyWithPosition(
                                isFirst: index == 0,
                                isLast: index == widget.itemCount - 1);
                          else
                            return widgetChild;
                        },
                      ),
              )
            : Container(),
      ],
    );
  }
}

class ItemSectionExtendedController {
  LinkedList<_ListenerEntry> _listeners = LinkedList<_ListenerEntry>();
  bool _extended;
  bool get extended => _extended;

  ItemSectionExtendedController({bool extended = true})
      : this._extended = extended;

  void addListener(void Function(bool extended) listener) {
    _listeners.add(_ListenerEntry(listener));
  }

  void removeListener(listener) {
    for (final _ListenerEntry entry in _listeners) {
      if (entry.listener == listener) {
        entry.unlink();
        return;
      }
    }
  }

  void setExtended(bool extended) {
    _extended = extended;
  }

  void setExtendedAndUpdate(bool extended) {
    _extended = extended;
    for (final _ListenerEntry entry in _listeners) {
      if (entry.list != null) entry.listener(_extended);
    }
  }
}

class _ListenerEntry extends LinkedListEntry<_ListenerEntry> {
  _ListenerEntry(this.listener);
  final void Function(bool extended) listener;
}

abstract class IListItem extends Widget {
  // ignore: missing_return
  IListItem copyWithPosition({bool isFirst, bool isLast}) {}

  @override
  // ignore: missing_return
  Element createElement() {}
}
